/*
 * mdrun.c - main code to run a md simulation
 *
 * author: Frank Zirkelbach <frank.zirkelbach@physik.uni-augsburg.de>
 *
 */

#include "mdrun.h"

/* potential */
#include "potentials/harmonic_oscillator.h"
#include "potentials/lennard_jones.h"
#include "potentials/albe.h"
#ifdef TERSOFF_ORIG
#include "potentials/tersoff_orig.h"
#else
#include "potentials/tersoff.h"
#endif

#define ME	"[mdrun]"

/*
 * parse function
 */

int mdrun_usage(void) {

	printf("%s usage:\n",ME);

	return 0;
}

int mdrun_parse_argv(t_mdrun *mdrun,int argc,char **argv) {

	int i;

	for(i=1;i<argc;i++) {

		if(argv[i][0]!='-') {
			printf("%s unknown switch: %s\n",ME,argv[i]);
			return -1;
		}

		switch(argv[i][1]) {
			case 'c':
				strncpy(mdrun->cfile,argv[++i],128);
				break;
			case 's':
				strncpy(mdrun->sdir,argv[++i],128);
				break;
			case 'h':
				mdrun_usage();
				break;
			default:
				printf("%s unknown option: %s\n",ME,argv[i]);
				mdrun_usage();
				return -1;
		}

	}

	return 0;
}

int del_stages(t_mdrun *mdrun) {

	t_list *sl;
	t_stage *stage;

	sl=&(mdrun->stage);

	list_reset_f(sl);

	if(sl->start==NULL)
		return 0;

	do {
		stage=sl->current->data;
		free(stage->params);
		free(stage);
	} while(list_next_f(sl)!=L_NO_NEXT_ELEMENT);

	return 0;
}

int add_stage(t_mdrun *mdrun,u8 type,void *params) {

	int psize;

	t_stage *stage;

	switch(type) {
		case STAGE_DISPLACE_ATOM:
			psize=sizeof(t_displace_atom_params);
			break;
		case STAGE_DEL_ATOMS:
			psize=sizeof(t_del_atoms_params);
			break;
		case STAGE_MODIFY_ATOMS:
			psize=sizeof(t_modify_atoms_params);
			break;
		case STAGE_INSERT_ATOMS:
			psize=sizeof(t_insert_atoms_params);
			break;
		case STAGE_INSERT_MIXED_ATOMS:
			psize=sizeof(t_insert_mixed_atoms_params);
			break;
		case STAGE_CONTINUE:
			psize=sizeof(t_continue_params);
			break;
		case STAGE_ANNEAL:
			psize=sizeof(t_anneal_params);
			break;
		case STAGE_CHAATTR:
			psize=sizeof(t_chaattr_params);
			break;
		case STAGE_CHSATTR:
			psize=sizeof(t_chsattr_params);
			break;
		case STAGE_SET_TEMP:
			psize=sizeof(t_set_temp_params);
			break;
		case STAGE_SET_TIMESTEP:
			psize=sizeof(t_set_timestep_params);
			break;
		case STAGE_FILL:
			psize=sizeof(t_fill_params);
			break;
		case STAGE_THERMAL_INIT:
			psize=0;
			break;
		default:
			printf("%s unknown stage type: %02x\n",ME,type);
			return -1;
	}

	stage=malloc(sizeof(t_stage));
	if(stage==NULL) {
		perror("[mdrun] malloc (add stage)");
		return -1;
	}

	stage->type=type;
	stage->executed=FALSE;

	stage->params=malloc(psize);
	if(stage->params==NULL) {
		perror("[mdrun] malloc (stage params)");
		return -1;
	}

	memcpy(stage->params,params,psize);

	list_add_immediate_f(&(mdrun->stage),stage);

	return 0;
}

int mdrun_parse_config(t_mdrun *mdrun) {

	int fd,ret;
	char error[128];
	char line[128];
	char *wptr;
	char word[32][64];
	int wcnt;
	int i,o;

	t_displace_atom_params dap;
	t_modify_atoms_params map;
	t_insert_atoms_params iap;
	t_insert_mixed_atoms_params imp;
	t_continue_params cp;
	t_anneal_params ap;
	t_chaattr_params cap;
	t_chsattr_params csp;
	t_set_temp_params stp;
	t_set_timestep_params stsp;
	t_fill_params fp;
	t_del_atoms_params delp;

	/* open config file */
	fd=open(mdrun->cfile,O_RDONLY);
	if(fd<0) {
		snprintf(error,128,"%s open cfile %s",ME,mdrun->cfile);
		perror(error);
		return fd;
	}

	/* read, parse, set */
	ret=1;
	while(ret>0) {

		/* read */
		ret=get_line(fd,line,128);

		/* parse */

		// ignore # lines and \n
		if((line[0]=='#')|(ret==1))
			continue;

		// reset
		memset(&iap,0,sizeof(t_insert_atoms_params));
		memset(&map,0,sizeof(t_modify_atoms_params));
		memset(&imp,0,sizeof(t_insert_mixed_atoms_params));
		memset(&cp,0,sizeof(t_continue_params));
		memset(&ap,0,sizeof(t_anneal_params));
		memset(&cap,0,sizeof(t_chaattr_params));
		memset(&csp,0,sizeof(t_chsattr_params));
		memset(&stp,0,sizeof(t_set_temp_params));
		memset(&stsp,0,sizeof(t_set_timestep_params));
		memset(&fp,0,sizeof(t_fill_params));
		memset(&delp,0,sizeof(t_del_atoms_params));

		// get command + args
		wcnt=0;
		while(1) {
			if(wcnt)
				wptr=strtok(NULL," \t");
			else
				wptr=strtok(line," \t");
			if(wptr==NULL)
				break;
			strncpy(word[wcnt],wptr,64);
			wcnt+=1;
		}

		if(!strncmp(word[0],"potential",9)) {
			if(!strncmp(word[1],"albe",4))
				mdrun->potential=MOLDYN_POTENTIAL_AM;
			if(!strncmp(word[1],"tersoff",7))
				mdrun->potential=MOLDYN_POTENTIAL_TM;
			if(!strncmp(word[1],"ho",2))
				mdrun->potential=MOLDYN_POTENTIAL_HO;
			if(!strncmp(word[1],"lj",2))
				mdrun->potential=MOLDYN_POTENTIAL_LJ;
		}
		else if(!strncmp(word[0],"continue",8))
			strncpy(mdrun->continue_file,word[1],128);
		else if(!strncmp(word[0],"cutoff",6))
			mdrun->cutoff=atof(word[1]);
		else if(!strncmp(word[0],"nnd",3))
			mdrun->nnd=atof(word[1]);
		else if(!strncmp(word[0],"intalgo",7)) {
			if(!strncmp(word[1],"verlet",5))
				mdrun->intalgo=MOLDYN_INTEGRATE_VERLET;
		}
		else if(!strncmp(word[0],"timestep",8))
			mdrun->timestep=atof(word[1]);
		else if(!strncmp(word[0],"volume",6)) {
			mdrun->dim.x=atof(word[1]);
			mdrun->dim.y=atof(word[2]);
			mdrun->dim.z=atof(word[3]);
			if(strncmp(word[4],"0",1))
				mdrun->vis=TRUE;
		}
		else if(!strncmp(word[0],"pbc",3)) {
			if(strncmp(word[1],"0",1))
				mdrun->pbcx=TRUE;
			else
				mdrun->pbcx=FALSE;
			if(strncmp(word[2],"0",1))
				mdrun->pbcy=TRUE;
			else
				mdrun->pbcy=FALSE;
			if(strncmp(word[3],"0",1))
				mdrun->pbcz=TRUE;
			else
				mdrun->pbcz=FALSE;
		}
		else if(!strncmp(word[0],"temperature",11))
			mdrun->temperature=atof(word[1]);
		else if(!strncmp(word[0],"pressure",8))
			mdrun->pressure=atof(word[1]);
		else if(!strncmp(word[0],"lattice",7)) {
			if(!strncmp(word[1],"zincblende",10))
				mdrun->lattice=ZINCBLENDE;
			if(!strncmp(word[1],"fcc",3))
				mdrun->lattice=FCC;
			if(!strncmp(word[1],"diamond",7))
				mdrun->lattice=DIAMOND;
			if(!strncmp(word[1],"none",4))
				mdrun->lattice=NONE;
			if(wcnt==3)
				mdrun->lc=atof(word[2]);
		}
		else if(!strncmp(word[0],"element1",8)) {
			mdrun->element1=atoi(word[1]);
		}
		else if(!strncmp(word[0],"element2",8)) {
			mdrun->element2=atoi(word[1]);
		}
		else if(!strncmp(word[0],"fill",6)) {
			// default values
			fp.fill_element=mdrun->element1;
			fp.fill_brand=0;
			fp.lattice=mdrun->lattice;
			fp.p_params.type=0;
			fp.d_params.type=0;
			fp.d_params.stype=0;
			// parse fill command
			i=1;
			while(i<wcnt) {
				if(!strncmp(word[i],"lc",2)) {
					fp.lx=atoi(word[++i]);
					fp.ly=atoi(word[++i]);
					fp.lz=atoi(word[++i]);
					fp.lc=atof(word[++i]);
					mdrun->lc=fp.lc;
					continue;
				}
				if(!strncmp(word[i],"eb",2)) {
					fp.fill_element=atoi(word[++i]);
					fp.fill_brand=atoi(word[++i]);
					continue;
				}
				if(word[i][0]=='p') {
					i+=1;
					switch(word[i][0]) {
					case 'i':
						if(word[i][1]=='r')
						fp.p_params.type=PART_INSIDE_R;
						else
						fp.p_params.type=PART_INSIDE_D;
						break;
					case 'o':
						if(word[i][1]=='r')
						fp.p_params.type=PART_OUTSIDE_R;
						else
						fp.p_params.type=PART_OUTSIDE_D;
						break;
					default:
						break;
					}
					if((fp.p_params.type==PART_INSIDE_R)||
                         		  (fp.p_params.type==PART_OUTSIDE_R)) {
						fp.p_params.r=atof(word[++i]);	
						fp.p_params.p.x=atof(word[++i]);
						fp.p_params.p.y=atof(word[++i]);
						fp.p_params.p.z=atof(word[++i]);
					}
					if((fp.p_params.type==PART_INSIDE_D)||
					   (fp.p_params.type==PART_OUTSIDE_D)) {
						fp.p_params.p.x=atof(word[++i]);
						fp.p_params.p.y=atof(word[++i]);
						fp.p_params.p.z=atof(word[++i]);
						fp.p_params.d.x=atof(word[++i]);
						fp.p_params.d.y=atof(word[++i]);
						fp.p_params.d.z=atof(word[++i]);
					}
					continue;
				}
				if(word[i][0]=='d') {
					switch(word[++i][0]) {
						case '0':

				fp.d_params.type=DEFECT_TYPE_0D;
				if(!strncmp(word[i+1],"dbx",3)) {
					fp.d_params.stype=DEFECT_STYPE_DB_X;
				}
				if(!strncmp(word[i+1],"dby",3)) {
					fp.d_params.stype=DEFECT_STYPE_DB_Y;
				}
				if(!strncmp(word[i+1],"dbz",3)) {
					fp.d_params.stype=DEFECT_STYPE_DB_Z;
				}
				if(!strncmp(word[i+1],"dbr",3)) {
					fp.d_params.stype=DEFECT_STYPE_DB_R;
				}
				i+=1;
				fp.d_params.od=atof(word[++i]);
				fp.d_params.dd=atof(word[++i]);
				fp.d_params.element=atoi(word[++i]);
				fp.d_params.brand=atoi(word[++i]);
				// parsed in future
		fp.d_params.attr=ATOM_ATTR_HB|ATOM_ATTR_VA;
		fp.d_params.attr|=ATOM_ATTR_1BP|ATOM_ATTR_2BP|ATOM_ATTR_3BP;
				break;

						case '1':
				fp.d_params.type=DEFECT_TYPE_1D;
				break;
						case '2':
				fp.d_params.type=DEFECT_TYPE_2D;
				break;
						case '3':
				fp.d_params.type=DEFECT_TYPE_3D;
				break;
						default:
							break;
					}
					continue;

				}
				// offset
				if(word[i][0]=='o') {
					fp.o_params.o.x=atof(word[++i])*fp.lc;
					fp.o_params.o.y=atof(word[++i])*fp.lc;
					fp.o_params.o.z=atof(word[++i])*fp.lc;
					fp.o_params.use=1;
					continue;
				}
				i+=1;
			}
			add_stage(mdrun,STAGE_FILL,&fp);
		}
		else if(!strncmp(word[0],"thermal_init",12)) {
			add_stage(mdrun,STAGE_THERMAL_INIT,NULL);
		}
		else if(!strncmp(word[0],"aattr",5)) {
			// for aatrib line we need a special stage
			// containing one schedule of 0 loops ...
			for(i=0;i<strlen(word[1]);i++) {
				switch(word[1][i]) {
					case 't':
						cap.type|=CHAATTR_TOTALV;
						break;
					case 'r':
						cap.type|=CHAATTR_REGION;
						break;
					case 'e':
						cap.type|=CHAATTR_ELEMENT;
						break;
					case 'n':
						cap.type|=CHAATTR_NUMBER;
					default:
						break;
				}
			}
			i=2;
			if(cap.type&CHAATTR_REGION) {
				cap.x0=atof(word[2]);	
				cap.y0=atof(word[3]);	
				cap.z0=atof(word[4]);	
				cap.x1=atof(word[5]);	
				cap.y1=atof(word[6]);	
				cap.z1=atof(word[7]);
				i+=6;
			}
			if(cap.type&CHAATTR_ELEMENT) {
				cap.element=atoi(word[i]);
				i+=1;
			}
			if(cap.type&CHAATTR_NUMBER) {
				cap.element=atoi(word[i]);
				i+=1;
			}
			for(o=0;o<strlen(word[i]);o++) {
				switch(word[i][o]) {
					case 'b':
						cap.attr|=ATOM_ATTR_VB;
						break;
					case 'h':
						cap.attr|=ATOM_ATTR_HB;
						break;
					case 'v':
						cap.attr|=ATOM_ATTR_VA;
						break;
					case 'f':
						cap.attr|=ATOM_ATTR_FP;
						break;
					case '1':
						cap.attr|=ATOM_ATTR_1BP;
						break;
					case '2':
						cap.attr|=ATOM_ATTR_2BP;
						break;
					case '3':
						cap.attr|=ATOM_ATTR_3BP;
						break;
					default:
						break;
				}
			}
			add_stage(mdrun,STAGE_CHAATTR,&cap);
		}
		else if(!strncmp(word[0],"sattr",5)) {
			// for satrib line we need a special stage
			// containing one schedule of 0 loops ...
			csp.type=0;
			for(i=1;i<wcnt;i++) {
				if(!strncmp(word[i],"pctrl",5)) {
					csp.ptau=atof(word[++i]);
					if(csp.ptau>0)
						csp.ptau=0.01/(csp.ptau*GPA);
					csp.type|=CHSATTR_PCTRL;
				}
				if(!strncmp(word[i],"tctrl",5)) {
					csp.ttau=atof(word[++i]);
					csp.type|=CHSATTR_TCTRL;
				}
				if(!strncmp(word[i],"prelax",6)) {
					csp.dp=atof(word[++i])*BAR;
					csp.type|=CHSATTR_PRELAX;
				}
				if(!strncmp(word[i],"trelax",6)) {
					csp.dt=atof(word[++i]);
					csp.type|=CHSATTR_TRELAX;
				}
				if(!strncmp(word[i],"rsteps",6)) {
					csp.rsteps=atoi(word[++i]);
					csp.type|=CHSATTR_RSTEPS;
				}
				if(!strncmp(word[i],"avgrst",6)) {
					csp.avgrst=atoi(word[++i]);
					csp.type|=CHSATTR_AVGRST;
				}
			}
			add_stage(mdrun,STAGE_CHSATTR,&csp);
		}
		else if(!strncmp(word[0],"prerun",6))
			mdrun->prerun=atoi(word[1]);
		else if(!strncmp(word[0],"avgskip",7))
			mdrun->avgskip=atoi(word[1]);
		else if(!strncmp(word[0],"elog",4))
			mdrun->elog=atoi(word[1]);
		else if(!strncmp(word[0],"tlog",4))
			mdrun->tlog=atoi(word[1]);
		else if(!strncmp(word[0],"plog",4))
			mdrun->plog=atoi(word[1]);
		else if(!strncmp(word[0],"vlog",4))
			mdrun->vlog=atoi(word[1]);
		else if(!strncmp(word[0],"save",4))
			mdrun->save=atoi(word[1]);
		else if(!strncmp(word[0],"visualize",9))
			mdrun->visualize=atoi(word[1]);
		else if(!strncmp(word[0],"stage",5)) {
			// for every stage line, add a stage
			if(!strncmp(word[1],"displace",8)) {
				dap.nr=atoi(word[2]);	
				dap.dx=atof(word[3]);
				dap.dy=atof(word[4]);
				dap.dz=atof(word[5]);
				add_stage(mdrun,STAGE_DISPLACE_ATOM,&dap);
			}
			else if(!strncmp(word[1],"del_atoms",9)) {
				delp.o.x=atof(word[2]);
				delp.o.y=atof(word[3]);
				delp.o.z=atof(word[4]);
				delp.r=atof(word[5]);
				add_stage(mdrun,STAGE_DEL_ATOMS,&delp);
			}
			else if(!strncmp(word[1],"mod_atoms",8)) {
				i=2;
				while(i<wcnt) {
					if(!strncmp(word[i],"t",1)) {
						map.tag=atoi(word[++i]);
						i+=1;
					}
					if(!strncmp(word[i],"ekin",5)) {
						map.ekin.x=atof(word[++i])*EV;
						map.ekin.y=atof(word[++i])*EV;
						map.ekin.z=atof(word[++i])*EV;
						i+=1;
					}
				}
				add_stage(mdrun,STAGE_MODIFY_ATOMS,&map);
			}
			else if(!strncmp(word[1],"ins_atoms",9)) {
				iap.ins_steps=atoi(word[2]);
				iap.ins_atoms=atoi(word[3]);
				iap.element=atoi(word[4]);
				iap.brand=atoi(word[5]);
				for(i=0;i<strlen(word[6]);i++) {
					switch(word[6][i]) {
						case 'b':
							iap.attr|=ATOM_ATTR_VB;
							break;
						case 'h':
							iap.attr|=ATOM_ATTR_HB;
							break;
						case 'v':
							iap.attr|=ATOM_ATTR_VA;
							break;
						case 'f':
							iap.attr|=ATOM_ATTR_FP;
							break;
						case '1':
							iap.attr|=ATOM_ATTR_1BP;
							break;
						case '2':
							iap.attr|=ATOM_ATTR_2BP;
							break;
						case '3':
							iap.attr|=ATOM_ATTR_3BP;
							break;
						default:
							break;
					}
				}
				switch(word[7][0]) {
					// insertion types
					case 'p':
						iap.type=INS_POS;
						iap.x0=atof(word[8]);
						iap.y0=atof(word[9]);
						iap.z0=atof(word[10]);
						break;
					case 'P':
						iap.type=INS_RELPOS;
						iap.x0=atof(word[8]);
						iap.y0=atof(word[9]);
						iap.z0=atof(word[10]);
						break;
					case 'r':
						switch(word[8][0]) {

						case 't':
							iap.type=INS_TOTAL;
							iap.cr=atof(word[9]);
							break;
						case 'r':
							iap.type=INS_RECT;
							iap.x0=atof(word[9]);
							iap.y0=atof(word[10]);
							iap.z0=atof(word[11]);
							iap.x1=atof(word[12]);
							iap.y1=atof(word[13]);
							iap.z1=atof(word[14]);
							iap.cr=atof(word[15]);
							break;
						case 's':
							iap.type=INS_SPHERE;
							iap.x0=atof(word[9]);
							iap.y0=atof(word[10]);
							iap.z0=atof(word[11]);
							iap.x1=atof(word[12]);
							iap.cr=atof(word[13]);
							break;
						default:
							break;
						}
					default:
						break;
				}
				add_stage(mdrun,STAGE_INSERT_ATOMS,&iap);
			}


			// HERE WE GO ...

			else if(!strncmp(word[1],"ins_m_atoms",11)) {
				imp.element1=atoi(word[2]);
				imp.element2=atoi(word[3]);
				imp.amount1=atoi(word[4]);
				imp.amount2=atoi(word[5]);
				imp.brand1=atoi(word[6]);
				imp.brand2=atoi(word[7]);
				imp.crmin=atof(word[8]);
				imp.crmax=atof(word[9]);
				/* do this later ...
				for(i=0;i<strlen(word[8]);i++) {
					switch(word[8][i]) {
						case 'b':
							imp.attr|=ATOM_ATTR_VB;
							break;
						case 'h':
							imp.attr|=ATOM_ATTR_HB;
							break;
						case 'v':
							imp.attr|=ATOM_ATTR_VA;
							break;
						case 'f':
							imp.attr|=ATOM_ATTR_FP;
							break;
						case '1':
							imp.attr|=ATOM_ATTR_1BP;
							break;
						case '2':
							imp.attr|=ATOM_ATTR_2BP;
							break;
						case '3':
							imp.attr|=ATOM_ATTR_3BP;
							break;
						default:
							break;
					}
				}
				*/
				imp.attr1=ATOM_ATTR_HB|ATOM_ATTR_VA|ATOM_ATTR_1BP|ATOM_ATTR_2BP|ATOM_ATTR_3BP|ATOM_ATTR_FP;
				imp.attr2=ATOM_ATTR_HB|ATOM_ATTR_VA|ATOM_ATTR_1BP|ATOM_ATTR_2BP|ATOM_ATTR_3BP|ATOM_ATTR_FP;
				add_stage(mdrun,STAGE_INSERT_MIXED_ATOMS,&imp);
			}





			else if(!strncmp(word[1],"continue",8)) {
				cp.runs=atoi(word[2]);
				add_stage(mdrun,STAGE_CONTINUE,&cp);
			}
			else if(!strncmp(word[1],"anneal",6)) {
				ap.count=0;
				ap.runs=atoi(word[2]);
				ap.dt=atof(word[3]);
				ap.interval=atoi(word[4]);
				add_stage(mdrun,STAGE_ANNEAL,&ap);
			}
			else if(!strncmp(word[1],"set_temp",8)) {
				if(word[2][0]=='c') {
					stp.type=SET_TEMP_CURRENT;
					stp.val=0.0;
				}
				else {
					stp.type=SET_TEMP_VALUE;
					stp.val=atof(word[2]);
				}
				add_stage(mdrun,STAGE_SET_TEMP,&stp);
			}
			else if(!strncmp(word[1],"set_timestep",12)) {
				stsp.tau=atof(word[2]);
				add_stage(mdrun,STAGE_SET_TIMESTEP,&stsp);
			}
			else {
				printf("%s unknown stage type: %s\n",
				       ME,word[1]);
				return -1;
			}
		}
		else {
			printf("%s unknown keyword '%s', skipped!\n",
			       ME,word[0]);
			continue;
		}
	}

	/* close file */
	close(fd);

	return 0;
}

int check_pressure(t_moldyn *moldyn,t_mdrun *mdrun) {

	double delta;

	if(!(mdrun->sattr&SATTR_PRELAX)) {
		return TRUE;
	}

	delta=moldyn->p_avg-moldyn->p_ref;

	if(delta<0)
		delta=-delta;

	if(delta>mdrun->dp)
		return FALSE;

	return TRUE;
}

int check_temperature(t_moldyn *moldyn,t_mdrun *mdrun) {

	double delta;

	if(!(mdrun->sattr&SATTR_TRELAX))
		return TRUE;

	delta=moldyn->t_avg-moldyn->t_ref;

	if(delta<0)
		delta=-delta;

	if(delta>mdrun->dt)
		return FALSE;

	return TRUE;
}

int displace_atom(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_displace_atom_params *dap;
	t_stage *stage;
	t_atom *atom;

	stage=mdrun->stage.current->data;
	dap=stage->params;

	atom=&(moldyn->atom[dap->nr]);
	atom->r.x+=dap->dx;
	atom->r.y+=dap->dy;
	atom->r.z+=dap->dz;

	return 0;
}

int del_atoms(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_stage *stage;
	t_del_atoms_params *delp;
	int i;
	t_3dvec dist;
	
	stage=mdrun->stage.current->data;
	delp=stage->params;

	for(i=0;i<moldyn->count;i++) {
		v3_sub(&dist,&(delp->o),&(moldyn->atom[i].r));
//printf("%d ----> %f %f %f = %f | %f\n",i,dist.x,dist.y,dist.z,v3_absolute_square(&dist),delp->r*delp->r);
		if(v3_absolute_square(&dist)<=(delp->r*delp->r)) {
			del_atom(moldyn,moldyn->atom[i].tag);
			printf("%s atom deleted: %d %d %d\n",ME,
                               moldyn->atom[i].tag,moldyn->atom[i].element,
			       moldyn->atom[i].brand);
		}
	}

	return 0;
}

int modify_atoms(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_modify_atoms_params *map;
	t_stage *stage;
	t_atom *atom;
	t_3dvec v;
	int i;

	atom=moldyn->atom;
	stage=mdrun->stage.current->data;
	map=stage->params;
	v.x=0.0; v.y=0.0; v.z=0.0;

	for(i=0;i<moldyn->count;i++) {
		if(atom[i].tag==map->tag) {
			v.x=sqrt(2.0*fabs(map->ekin.x)/atom[i].mass);
			if(map->ekin.x<0.0)
				v.x=-v.x;
			v.y=sqrt(2.0*fabs(map->ekin.y)/atom[i].mass);
			if(map->ekin.y<0.0)
				v.y=-v.y;
			v.z=sqrt(2.0*fabs(map->ekin.z)/atom[i].mass);
			if(map->ekin.z<0.0)
				v.z=-v.z;
			v3_copy(&(atom[i].v),&v);
			printf("%s atom modified: v = (%f %f %f)\n",
                               ME,v.x,v.y,v.z);
		}
	}	

	return 0;
}

int insert_atoms(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_insert_atoms_params *iap;
	t_stage *stage;
	t_atom *atom;
	t_3dvec r,v,dist;
	double d,dmin,o;
	int cnt,i;
	double x,y,z;
	double x0,y0,z0;
	u8 cr_check,run;
	
	stage=mdrun->stage.current->data;
	iap=stage->params;

	cr_check=FALSE;

	v.x=0.0; v.y=0.0; v.z=0.0;
	x=0; y=0; z=0;
	o=0; dmin=0;

	switch(mdrun->lattice) {
		case CUBIC:
			o=0.5*mdrun->lc;
			break;
		case FCC:
			o=0.25*mdrun->lc;
			break;
		case DIAMOND:
		case ZINCBLENDE:
			o=0.125*mdrun->lc;
			break;
		default:
			break;
	}

	switch(iap->type) {
		case INS_TOTAL:
			x=moldyn->dim.x;
			x0=-x/2.0;
			y=moldyn->dim.y;
			y0=-y/2.0;
			z=moldyn->dim.z;
			z0=-z/2.0;
			cr_check=TRUE;
			break;
		case INS_RECT:
			x=iap->x1-iap->x0;
			x0=iap->x0;
			y=iap->y1-iap->y0;
			y0=iap->y0;
			z=iap->z1-iap->z0;
			z0=iap->z0;
			cr_check=TRUE;
			break;
		case INS_SPHERE:
			x=2.0*iap->x1;
			x0=iap->x0-iap->x1;
			y=x;
			y0=iap->y0-iap->x1;
			z=x;
			z0=iap->z0-iap->x1;
			cr_check=TRUE;
			break;
		case INS_POS:
		case INS_RELPOS:
			x0=iap->x0;
			y0=iap->y0;
			z0=iap->z0;
			cr_check=FALSE;
			break;
		default:
			printf("%s unknown insertion mode: %02x\n",
			       ME,iap->type);
			return -1;
	}

	cnt=0;
	while(cnt<iap->ins_atoms) {
		run=1;
		while(run) {
			if((iap->type!=INS_POS)&&(iap->type!=INS_RELPOS)) {
				r.x=rand_get_double(&(moldyn->random))*x;
				r.y=rand_get_double(&(moldyn->random))*y;
				r.z=rand_get_double(&(moldyn->random))*z;
			}
			else {
				r.x=0.0;
				r.y=0.0;
				r.z=0.0;
			}
			if(iap->type==INS_RELPOS) {
				r.x+=x0*mdrun->lc;
				r.y+=y0*mdrun->lc;
				r.z+=z0*mdrun->lc;
			}
			else {
				r.x+=x0;
				r.y+=y0;
				r.z+=z0;
			}
			// offset
			if(iap->type!=INS_TOTAL) {
				r.x+=o;
				r.y+=o;
				r.z+=o;
			}
			run=0;
			if(cr_check==TRUE) {
				dmin=1000;	// for sure too high!
				for(i=0;i<moldyn->count;i++) {
					atom=&(moldyn->atom[i]);
					v3_sub(&dist,&(atom->r),&r);
					check_per_bound(moldyn,&dist);
					d=v3_absolute_square(&dist);
					if(d<(iap->cr*iap->cr)) {
						run=1;
						break;
					}
					if(d<dmin)
						dmin=d;
				}
			}
			if(iap->type==INS_SPHERE) {
				if((r.x-iap->x0)*(r.x-iap->x0)+
				   (r.y-iap->y0)*(r.y-iap->y0)+
				   (r.z-iap->z0)*(r.z-iap->z0)>
				   (iap->x1*iap->x1)) {
					run=1;
				}
			}
		}
		add_atom(moldyn,iap->element,
		         iap->brand,iap->attr,&r,&v);
		printf("%s atom inserted (%d/%d): %f %f %f\n",
		       ME,(iap->cnt_steps+1)*iap->ins_atoms,
		       iap->ins_steps*iap->ins_atoms,r.x,r.y,r.z);
		printf("  attributes: ");
		if(iap->attr&ATOM_ATTR_VB)
			printf("b ");
		if(iap->attr&ATOM_ATTR_HB)
			printf("h ");
		if(iap->attr&ATOM_ATTR_VA)
			printf("v ");
		if(iap->attr&ATOM_ATTR_FP)
			printf("f ");
		if(iap->attr&ATOM_ATTR_1BP)
			printf("1 ");
		if(iap->attr&ATOM_ATTR_2BP)
			printf("2 ");
		if(iap->attr&ATOM_ATTR_3BP)
			printf("3 ");
		printf("\n");
		printf("  d2 = %f/%f\n",dmin,iap->cr*iap->cr);
		cnt+=1;
	}

	return 0;
}

int insert_mixed_atoms(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_stage *stage;
	t_insert_mixed_atoms_params *imp;
	t_atom *atom;
	double x,x0,y,y0,z,z0;
	double dmin,d,cmin,cmax;
	u8 retry;
	t_3dvec r,v,dist;
	int i;
	

	stage=mdrun->stage.current->data;
	imp=stage->params;

	x=moldyn->dim.x;
	x0=-x/2.0;
	y=moldyn->dim.y;
	y0=-y/2.0;
	z=moldyn->dim.z;
	z0=-z/2.0;

	v.x=0.0;
	v.y=0.0;
	v.z=0.0;

	cmin=imp->crmin*imp->crmin;
	cmax=imp->crmax*imp->crmax;

	while(imp->amount1|imp->amount2) {
		if(imp->amount1) {
			retry=1;
			while(retry) {
				retry=0;
				r.x=rand_get_double(&(moldyn->random))*x;
				r.y=rand_get_double(&(moldyn->random))*y;
				r.z=rand_get_double(&(moldyn->random))*z;
				r.x+=x0;
				r.y+=y0;
				r.z+=z0;
				dmin=1000.0;	// for sure too high!
				for(i=0;i<moldyn->count;i++) {
					atom=&(moldyn->atom[i]);
					v3_sub(&dist,&(atom->r),&r);
					check_per_bound(moldyn,&dist);
					d=v3_absolute_square(&dist);
					if(d<cmin) {
						retry=1;
						break;
					}
					if(d<dmin)
						dmin=d;
				}
				if(dmin!=1000.0)
					if(dmin>cmax)
						retry=1;
			}
			add_atom(moldyn,imp->element1,
                                 imp->brand1,imp->attr1,&r,&v);
			printf("%s (mixed) atom inserted (%d): %f %f %f\n",
                               ME,imp->amount1,r.x,r.y,r.z);
			printf("  -> d2 = %f/%f/%f\n",dmin,cmin,cmax);
			imp->amount1-=1;
		}
		if(imp->amount2) {
			retry=1;
			while(retry) {
				retry=0;
				r.x=rand_get_double(&(moldyn->random))*x;
				r.y=rand_get_double(&(moldyn->random))*y;
				r.z=rand_get_double(&(moldyn->random))*z;
				r.x+=x0;
				r.y+=y0;
				r.z+=z0;
				dmin=1000.0;	// for sure too high!
				for(i=0;i<moldyn->count;i++) {
					atom=&(moldyn->atom[i]);
					v3_sub(&dist,&(atom->r),&r);
					check_per_bound(moldyn,&dist);
					d=v3_absolute_square(&dist);
					if(d<cmin) {
						retry=1;
						break;
					}
					if(d<dmin)
						dmin=d;
				}
				if(dmin!=1000.0)
					if(dmin>cmax)
						retry=1;
			}
			add_atom(moldyn,imp->element2,
                                 imp->brand2,imp->attr2,&r,&v);
			printf("%s (mixed) atom inserted (%d): %f %f %f\n",
                               ME,imp->amount2,r.x,r.y,r.z);
			printf("  -> d2 = %f/%f/%f\n",dmin,cmin,cmax);
			imp->amount2-=1;
		}
	}

	return 0;
}

int chaatr(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_stage *stage;
	t_chaattr_params *cap;
	t_atom *atom;
	int i;

	stage=mdrun->stage.current->data;
	cap=stage->params;

	for(i=0;i<moldyn->count;i++) {
		atom=&(moldyn->atom[i]);
		if(cap->type&CHAATTR_ELEMENT) {
			if(cap->element!=atom->element)
				continue;
		}
		if(cap->type&CHAATTR_NUMBER) {
			if(cap->element!=atom->tag)
				continue;
		}
		if(cap->type&CHAATTR_REGION) {
			if(cap->x0>atom->r.x)
				continue;
			if(cap->y0>atom->r.y)
				continue;
			if(cap->z0>atom->r.z)
				continue;
			if(cap->x1<atom->r.x)
				continue;
			if(cap->y1<atom->r.y)
				continue;
			if(cap->z1<atom->r.z)
				continue;
		}
		if(!(cap->type&CHAATTR_TOTALV))
			printf("  changing attributes of atom %d (0x%x)\n",
			       i,cap->attr);
		atom->attr=cap->attr;
	}

	printf("\n\n");

	return 0;
}

int chsattr(t_moldyn *moldyn,t_mdrun *mdrun) {

	t_stage *stage;
	t_chsattr_params *csp;

	stage=mdrun->stage.current->data;
	csp=stage->params;

	if(csp->type&CHSATTR_PCTRL) {
		if(csp->ptau>0)
			set_p_scale(moldyn,P_SCALE_BERENDSEN,csp->ptau);
		else
			set_p_scale(moldyn,P_SCALE_NONE,1.0);
	}
	if(csp->type&CHSATTR_TCTRL) {
		if(csp->ttau>0)
			set_t_scale(moldyn,T_SCALE_BERENDSEN,csp->ttau);
		else
			set_t_scale(moldyn,T_SCALE_NONE,1.0);
	}
	if(csp->type&CHSATTR_PRELAX) {
		if(csp->dp<0)
			mdrun->sattr&=(~(SATTR_PRELAX));
		else
			mdrun->sattr|=SATTR_PRELAX;
		mdrun->dp=csp->dp;
	}
	if(csp->type&CHSATTR_TRELAX) {
		if(csp->dt<0)
			mdrun->sattr&=(~(SATTR_TRELAX));
		else
			mdrun->sattr|=SATTR_TRELAX;
		mdrun->dt=csp->dt;
	}
	if(csp->type&CHSATTR_AVGRST) {
		if(csp->avgrst)
			mdrun->sattr|=SATTR_AVGRST;
		else
			mdrun->sattr&=(~(SATTR_AVGRST));
	}
	if(csp->type&CHSATTR_RSTEPS) {
		mdrun->relax_steps=csp->rsteps;
	}

	return 0;
}

#define stage_print(m)	if(!(stage->executed)) \
				printf("%s",m)

int mdrun_hook(void *ptr1,void *ptr2) {

	t_moldyn *moldyn;
	t_mdrun *mdrun;
	t_stage *stage;
	t_list *sl;
	int steps;
	u8 change_stage;
	t_3dvec o;

	t_insert_atoms_params *iap;
	t_insert_mixed_atoms_params *imp;
	t_continue_params *cp;
	t_anneal_params *ap;
	t_set_temp_params *stp;
	t_set_timestep_params *stsp;
	t_fill_params *fp;

	moldyn=ptr1;
	mdrun=ptr2;

	sl=&(mdrun->stage);

	change_stage=FALSE;

	/* return immediately if there are no more stage */
	if(sl->current==NULL)
		return 0;

	/* get stage description */
	stage=sl->current->data;

	/* steps in next schedule */
	steps=mdrun->relax_steps;

	/* check whether relaxation steps are necessary */
	if((check_pressure(moldyn,mdrun)==TRUE)&\
	   (check_temperature(moldyn,mdrun)==TRUE)) {
	
		/* be verbose */
		stage_print("\n###########################\n");
		stage_print("# [mdrun] executing stage #\n");
		stage_print("###########################\n\n");
		
		/* stage specific stuff */
		switch(stage->type) {
			case STAGE_DISPLACE_ATOM:
				stage_print("  -> displace atom\n\n");
				displace_atom(moldyn,mdrun);
				change_stage=TRUE;
				break;
			case STAGE_DEL_ATOMS:
				stage_print(" -> del atoms\n\n");
				del_atoms(moldyn,mdrun);
				change_stage=TRUE;
				break;
			case STAGE_MODIFY_ATOMS:
				stage_print(" -> modify atoms\n\n");
				modify_atoms(moldyn,mdrun);
				change_stage=TRUE;
				break;
			case STAGE_INSERT_ATOMS:
				stage_print("  -> insert atoms\n\n");
				iap=stage->params;
				if(iap->cnt_steps==iap->ins_steps) {
					change_stage=TRUE;
					break;
				}
				insert_atoms(moldyn,mdrun);
				iap->cnt_steps+=1;
				break;



			case STAGE_INSERT_MIXED_ATOMS:
				stage_print("  -> insert mixed atoms\n\n");
				imp=stage->params;
				insert_mixed_atoms(moldyn,mdrun);
				change_stage=TRUE;
				break;



			case STAGE_CONTINUE:
				stage_print("  -> continue\n\n");
				if(stage->executed==TRUE) {
					change_stage=TRUE;
					break;
				}
				cp=stage->params;
				steps=cp->runs;
				break;
			case STAGE_ANNEAL:
				stage_print("  -> anneal\n\n");
				ap=stage->params;
				if(ap->count==ap->runs) {
					change_stage=TRUE;
					break;
				}
				if(moldyn->t_ref+ap->dt>=0.0)
					set_temperature(moldyn,
					                moldyn->t_ref+ap->dt);
				ap->count+=1;
				steps=ap->interval;
				break;
			case STAGE_CHAATTR:
				stage_print("  -> change atom attributes\n\n");
				chaatr(moldyn,mdrun);
				change_stage=TRUE;
				break;
			case STAGE_CHSATTR:
				stage_print("  -> change sys attributes\n\n");
				chsattr(moldyn,mdrun);
				change_stage=TRUE;
				break;
			case STAGE_SET_TEMP:
				stage_print("  -> set temperature\n\n");
				stp=stage->params;
				if(stp->type&SET_TEMP_CURRENT) {
					set_temperature(moldyn,moldyn->t_avg);
				}
				else {
					set_temperature(moldyn,stp->val);
				}
				change_stage=TRUE;
				break;
			case STAGE_SET_TIMESTEP:
				stage_print("  -> set timestep\n\n");
				stsp=stage->params;
				mdrun->timestep=stsp->tau;
				change_stage=TRUE;
				break;
			case STAGE_FILL:
				stage_print("  -> fill lattice\n\n");
				fp=stage->params;
				switch(fp->lattice) {
					case ZINCBLENDE:

					o.x=0.5*0.25*fp->lc;
					o.y=o.x;
					o.z=o.x;
					create_lattice(moldyn,
					               FCC,fp->lc,
					               mdrun->element1,
					               DEFAULT_ATOM_ATTR,0,
					               fp->lx,fp->ly,fp->lz,
					               &o,
					               &(fp->p_params),
					               &(fp->d_params),
					               &(fp->o_params));
					o.x+=0.25*fp->lc;
					o.y=o.x;
					o.z=o.x;
					create_lattice(moldyn,
					               FCC,fp->lc,
					               mdrun->element2,
					               DEFAULT_ATOM_ATTR,1,
					               fp->lx,fp->ly,fp->lz,
					               &o,
					               &(fp->p_params),
					               &(fp->d_params),
					               &(fp->o_params));
					break;

					default:

					create_lattice(moldyn,
					               fp->lattice,fp->lc,
					               fp->fill_element,
					               DEFAULT_ATOM_ATTR,
					               fp->fill_brand,
					               fp->lx,fp->ly,fp->lz,
					               NULL,
					               &(fp->p_params),
					               &(fp->d_params),
					               &(fp->o_params));
					break;
				}
				moldyn_bc_check(moldyn);
				change_stage=TRUE;
				break;
			case STAGE_THERMAL_INIT:
				stage_print("  -> thermal init\n\n");
				thermal_init(moldyn,TRUE);
				change_stage=TRUE;
				break;
			default:
				printf("%s unknwon stage type\n",ME);
				break;
		}
	
		/* mark as executed */
		stage->executed=TRUE;
	
		/* change state */
		if(change_stage==TRUE) {
			printf("%s finished stage\n",ME);
			if(list_next_f(sl)==L_NO_NEXT_ELEMENT) {
				printf("%s no more stages\n",ME);
				return 0;
			}
			steps=0;
		}

	}
	else {

		/* averages */
		if(mdrun->sattr&SATTR_AVGRST)
			average_reset(moldyn);

	}

	/* continue simulation */
	moldyn_add_schedule(moldyn,steps,mdrun->timestep);

	return 0;
}

int main(int argc,char **argv) {

	t_mdrun mdrun;
	t_moldyn moldyn;
	//t_3dvec o;

	/* clear structs */
	memset(&mdrun,0,sizeof(t_mdrun));
	memset(&moldyn,0,sizeof(t_moldyn));

	/* parse arguments */
	if(mdrun_parse_argv(&mdrun,argc,argv)<0)
		return -1;

	/* initialize list system */
	list_init_f(&(mdrun.stage));

	/* parse config file */
	mdrun_parse_config(&mdrun);

	/* reset the stage list */
	list_reset_f(&(mdrun.stage));

	/* sanity check! */

	/* prepare simulation */

	if(mdrun.continue_file[0]) {
		// read the save file
		moldyn_read_save_file(&moldyn,mdrun.continue_file);
		// manualaadjusting some stuff
		moldyn.argc=argc;
		moldyn.args=argv;
		rand_init(&(moldyn.random),NULL,1);
		moldyn.random.status|=RAND_STAT_VERBOSE;
	}
	else {
		moldyn_init(&moldyn,argc,argv);
	}
	
	if(set_int_alg(&moldyn,mdrun.intalgo)<0)
		return -1;

	/* potential */
	set_cutoff(&moldyn,mdrun.cutoff);
	if(set_potential(&moldyn,mdrun.potential)<0)
		return -1;
	switch(mdrun.potential) {
		case MOLDYN_POTENTIAL_AM:
			albe_mult_set_params(&moldyn,
			                     mdrun.element1,
			                     mdrun.element2);
			break;
		case MOLDYN_POTENTIAL_TM:
			tersoff_mult_set_params(&moldyn,
			                        mdrun.element1,
			                        mdrun.element2);
			break;
		case MOLDYN_POTENTIAL_HO:
			harmonic_oscillator_set_params(&moldyn,mdrun.element1);
			break;
		case MOLDYN_POTENTIAL_LJ:
			lennard_jones_set_params(&moldyn,mdrun.element1);
			break;
		default:
			printf("%s unknown potential: %02x\n",
			       ME,mdrun.potential);
			return -1;
	}

	/* if it is a continue run, reset schedule and skip lattice init */
	if(mdrun.continue_file[0]) {
		memset(&(moldyn.schedule),0,sizeof(t_moldyn_schedule));
		goto addsched;
	}

	/* initial lattice and dimensions */
	set_dim(&moldyn,mdrun.dim.x,mdrun.dim.y,mdrun.dim.z,mdrun.vis);
	set_pbc(&moldyn,mdrun.pbcx,mdrun.pbcy,mdrun.pbcz);
	/* replaced by fill stage !! 
	switch(mdrun.lattice) {
		case FCC:
			create_lattice(&moldyn,FCC,mdrun.lc,mdrun.fill_element,
			               mdrun.m1,DEFAULT_ATOM_ATTR,
                                       mdrun.fill_brand,mdrun.lx,
			               mdrun.ly,mdrun.lz,NULL,0,NULL);
			break;
		case DIAMOND:
			create_lattice(&moldyn,DIAMOND,mdrun.lc,
                                       mdrun.fill_element,
			               mdrun.m1,DEFAULT_ATOM_ATTR,
                                       mdrun.fill_brand,mdrun.lx,
			               mdrun.ly,mdrun.lz,NULL,0,NULL);
			break;
		case ZINCBLENDE:
			o.x=0.5*0.25*mdrun.lc; o.y=o.x; o.z=o.x;
			create_lattice(&moldyn,FCC,mdrun.lc,mdrun.element1,
			               mdrun.m1,DEFAULT_ATOM_ATTR,0,mdrun.lx,
			               mdrun.ly,mdrun.lz,&o,0,NULL);
			o.x+=0.25*mdrun.lc; o.y=o.x; o.z=o.x;
			create_lattice(&moldyn,FCC,mdrun.lc,mdrun.element2,
			               mdrun.m2,DEFAULT_ATOM_ATTR,1,mdrun.lx,
			               mdrun.ly,mdrun.lz,&o,0,NULL);
			break;
		case NONE:
			set_nn_dist(&moldyn,mdrun.nnd);
			break;
		default:
			printf("%s unknown lattice type: %02x\n",
			       ME,mdrun.lattice);
			return -1;
	}
	moldyn_bc_check(&moldyn);
	replaced by fill stage !! */

	/* temperature and pressure */
	set_temperature(&moldyn,mdrun.temperature);
	set_pressure(&moldyn,mdrun.pressure);
	/* replaced thermal_init stage
	thermal_init(&moldyn,TRUE);
	*/

addsched:
	/* first schedule */
	moldyn_add_schedule(&moldyn,mdrun.prerun,mdrun.timestep);

	/* log */
	moldyn_set_log_dir(&moldyn,mdrun.sdir);
	moldyn_set_report(&moldyn,"CHANGE ME","CHANGE ME TOO");
	if(mdrun.elog)
		moldyn_set_log(&moldyn,LOG_TOTAL_ENERGY,mdrun.elog);
	if(mdrun.tlog)
		moldyn_set_log(&moldyn,LOG_TEMPERATURE,mdrun.tlog);
	if(mdrun.plog)
		moldyn_set_log(&moldyn,LOG_PRESSURE,mdrun.plog);
	if(mdrun.vlog)
		moldyn_set_log(&moldyn,LOG_VOLUME,mdrun.vlog);
	if(mdrun.visualize)
		moldyn_set_log(&moldyn,VISUAL_STEP,mdrun.visualize);
	if(mdrun.save)
		moldyn_set_log(&moldyn,SAVE_STEP,mdrun.save);
	moldyn_set_log(&moldyn,CREATE_REPORT,0);
	set_avg_skip(&moldyn,mdrun.avgskip);

	/* prepare the hook function */
	moldyn_set_schedule_hook(&moldyn,&mdrun_hook,&mdrun);

	/* run the simulation */
	moldyn_integrate(&moldyn);

	/* shutdown */
	moldyn_shutdown(&moldyn);
	del_stages(&mdrun);
	list_destroy_f(&(mdrun.stage));

	return 0;
}
